﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Windows.Forms.VisualStyles;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment.Views {
  public partial class ToolTipComboBox : ComboBox {

    private ComboBoxState comboBoxState = ComboBoxState.Normal;
    private bool mouseOver = false;

    public ToolTipComboBox()
      : base() {
      SetStyle(ControlStyles.AllPaintingInWmPaint, true);
      SetStyle(ControlStyles.UserPaint, true);
      InitializeComponent();
      DrawMode = System.Windows.Forms.DrawMode.OwnerDrawVariable;
      DropDownStyle = ComboBoxStyle.DropDownList;
    }

    protected override void OnMeasureItem(MeasureItemEventArgs e) {
      base.OnMeasureItem(e);
      string text = GetItemText(Items[e.Index]);
      if (string.IsNullOrEmpty(text)) text = "test";
      var rect = TextRenderer.MeasureText(text, Font);
      e.ItemHeight = rect.Height - 2;
      e.ItemWidth = rect.Width;
    }

    protected override void OnDrawItem(DrawItemEventArgs e) {
      base.OnDrawItem(e);
      var args = new ToolTipRequiredEventArgs(e.Index, Items[e.Index]);
      OnToolTipRequired(args);
      if (!e.State.HasFlag(DrawItemState.ComboBoxEdit) && !string.IsNullOrEmpty(args.ToolTip)) {
        var ttRect = TextRenderer.MeasureText(args.ToolTip, Font);
        if (RectangleToScreen(e.Bounds).Right + ttRect.Width > Screen.FromControl(this).Bounds.Right)
          toolTip.Show(args.ToolTip, this, e.Bounds.Left - ttRect.Width, e.Bounds.Bottom);
        else toolTip.Show(args.ToolTip, this, e.Bounds.Right, e.Bounds.Bottom);
      } else toolTip.Hide(this);
      if (!e.State.HasFlag(DrawItemState.ComboBoxEdit))
        e.DrawBackground();
      if (e.Index > -1 && Items.Count > 0) {
        if (!e.State.HasFlag(DrawItemState.ComboBoxEdit) && e.State.HasFlag(DrawItemState.Selected)) {
          using (var b = new SolidBrush(BackColor)) {
            e.Graphics.DrawString(GetItemText(Items[e.Index]), Font, b, e.Bounds);
          }
        } else if (!e.State.HasFlag(DrawItemState.ComboBoxEdit)) {
          using (var b = new SolidBrush(ForeColor)) {
            e.Graphics.DrawString(GetItemText(Items[e.Index]), Font, b, e.Bounds);
          }
        } else {
          using (var b = new SolidBrush(ForeColor)) {
            Rectangle r = e.Bounds;
            r.Offset(0, 1);
            e.Graphics.DrawString(GetItemText(Items[e.Index]), Font, b, r);
          }
        }
      }
      e.DrawFocusRectangle();
    }

    protected override void OnPaint(PaintEventArgs e) {
      base.OnPaint(e);
      if (Enabled == false) comboBoxState = ComboBoxState.Disabled;
      Rectangle clip = Bounds; // using e.ClipRectangle leads to strange behavior when maximizing/unmaximizing, just redraw the entire region all the time
      clip.Offset(-clip.X, -clip.Y); // x and y need to be 0
      if (VisualStyleInformation.IsSupportedByOS && VisualStyleInformation.IsEnabledByUser) {
        PushButtonState pb_State = PushButtonState.Normal;
        if (comboBoxState == ComboBoxState.Disabled)
          pb_State = PushButtonState.Disabled;
        else if (comboBoxState == ComboBoxState.Hot)
          pb_State = PushButtonState.Hot;
        else if (comboBoxState == ComboBoxState.Pressed)
          pb_State = PushButtonState.Pressed;

        ButtonRenderer.DrawButton(e.Graphics, clip, pb_State);
        var r = new Rectangle(clip.Right - SystemInformation.VerticalScrollBarWidth + 2,
                                    clip.Top,
                                    SystemInformation.VerticalScrollBarWidth - 4,
                                    clip.Height);
        DrawArrow(e.Graphics, r);
      } else {
        ComboBoxRenderer.DrawTextBox(e.Graphics, clip, comboBoxState);
        var r = new Rectangle(clip.Right - SystemInformation.VerticalScrollBarWidth,
                                    clip.Top,
                                    SystemInformation.VerticalScrollBarWidth,
                                    clip.Height);
        ComboBoxRenderer.DrawDropDownButton(e.Graphics, r, comboBoxState);
      }
      Rectangle rect = clip;
      rect.Inflate(-2, -4);
      rect.Offset(1, 0);
      if (SelectedIndex > -1 && Items.Count > 0) {
        using (var b = new SolidBrush(ForeColor)) {
          e.Graphics.DrawString(GetItemText(Items[SelectedIndex]), Font, b, rect);
        }
      }
    }

    private void DrawArrow(Graphics g, Rectangle rect) {
      Point p1 = new Point(rect.X + 3, rect.Y + rect.Height / 2 - 2);
      Point p2 = new Point(rect.X + rect.Width - 3, rect.Y + rect.Height / 2 - 2);
      Point p3 = new Point(rect.X + rect.Width / 2, rect.Y + rect.Height / 2 + 2);
      g.FillPolygon(new SolidBrush(System.Drawing.SystemColors.ControlText), new Point[] { p1, p2, p3 });
    }

    protected override void OnDropDown(EventArgs e) {
      base.OnDropDown(e);
      comboBoxState = ComboBoxState.Pressed;
      Invalidate();
    }

    protected override void OnDropDownClosed(EventArgs e) {
      base.OnDropDownClosed(e);
      toolTip.Hide(this);
      comboBoxState = ComboBoxState.Normal;
      Invalidate();
    }

    protected override void OnMouseEnter(EventArgs e) {
      base.OnMouseEnter(e);
      mouseOver = true;
      if (comboBoxState == ComboBoxState.Normal) {
        comboBoxState = ComboBoxState.Hot;
        Invalidate();
      }
    }

    protected override void OnMouseHover(EventArgs e) {
      base.OnMouseHover(e);
      if (comboBoxState == ComboBoxState.Normal) {
        comboBoxState = ComboBoxState.Hot;
        Invalidate();
      }
    }

    protected override void OnMouseLeave(EventArgs e) {
      base.OnMouseLeave(e);
      mouseOver = false;
      if (comboBoxState == ComboBoxState.Hot) {
        comboBoxState = ComboBoxState.Normal;
        Invalidate();
      }
    }

    protected override void OnMouseDown(MouseEventArgs e) {
      base.OnMouseDown(e);
      comboBoxState = ComboBoxState.Pressed;
      Invalidate();
    }

    public event EventHandler<ToolTipRequiredEventArgs> ToolTipRequired;
    protected virtual void OnToolTipRequired(ToolTipRequiredEventArgs args) {
      var handler = ToolTipRequired;
      if (handler != null) handler(this, args);
    }
  }
}
