﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


namespace HeuristicLab.Problems.Instances {
  public enum CVRPDistanceMeasure { Direct, Euclidean, RoundedEuclidean, UpperEuclidean, Geo, Manhattan, Maximum, Att };

  /// <summary>
  /// Describes instances of the Capacitated Vehicle Routing Problem (CVRP).
  /// </summary>
  public interface ICVRPInstance {
    /// <summary>
    /// The name of the instance
    /// </summary>
    string Name { get; }
    /// <summary>
    /// Optional! The description of the instance
    /// </summary>
    string Description { get; }

    /// <summary>
    /// The number of customers and the depot
    /// </summary>
    int Dimension { get; }
    /// <summary>
    /// The distance measure that is used to calculate the distance between
    ///the coordinates if no <see cref="Distances"/> is given.
    /// </summary>
    CVRPDistanceMeasure DistanceMeasure { get; }
    /// <summary>
    /// Optional! The maximum number of vehicles that can be used.
    /// </summary>
    /// <remarks>
    /// If no number is given, but a maximum is required, it can be assumed that
    /// the maximum number of vehicles is equal to the number of customers as
    /// there cannot be more than one vehicle per customer.
    /// </remarks>
    double? MaximumVehicles { get; }
    /// <summary>
    /// The capacity of the vehicles, which is the same for all (homogeneous fleet).
    /// </summary>
    double Capacity { get; }
    /// <summary>
    /// Optional! The distances are given in form of a distance matrix.
    /// </summary>
    /// <remarks>
    /// Either Distances or the <see cref="Coordinates"/> need to be specified along
    /// with a distance measure.
    /// </remarks>
    double[,] Distances { get; }
    /// <summary>
    /// Optional! A a matrix of dimension [N, 2] where each row is either the customer
    /// or the depot and the columns represent x and y coordinates respectively.
    /// </summary>
    /// <remarks>
    /// Either <see cref="Distances"/> or the Coordinates need to be specified along
    /// with a distance measure.
    /// </remarks>
    double[,] Coordinates { get; }
    /// <summary>
    /// The demand vector that specifies how many goods need to be delivered.
    /// The vector has to include the depot, but with a demand of 0.
    /// </summary>
    double[] Demands { get; }

    /// <summary>
    /// Optional! The best-known solution as a list of tours in path-encoding.
    /// </summary>
    int[][] BestKnownTour { get; }
    /// <summary>
    /// Optional! The quality of the best-known solution.
    /// </summary>
    double? BestKnownQuality { get; }
  }
}
